/* $Id$ */
/*
 * Copyright (c) 2003-2004 Jose Nazario <jose@monkey.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Jose Nazario.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include <ctype.h>
#include <err.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "distance.h"

#define DAMERAU_D	0x0001
#define LEVENSHTEIN_D	0x0002
#define HAMMING_D	0x0004

const char rcsid[] = "Copyright (c) 2004 Jose Nazario, all rights reserved.";
const char version[] = "0.2.0";

void
usage(char *progname)
{
	fprintf(stderr, "Usage: %s [-DHLins] -d dist needle [file [...]]\n", 
		progname);
	fprintf(stderr, "search for similar text using fuzzy rules.\n");
	fprintf(stderr, "\t-D 	use the Damerau distance algorithm\n");
	fprintf(stderr, "\t-H 	use the Hamming distance algorithm\n");
	fprintf(stderr, "\t-L 	use the Levenshtein distance algorithm\n");
	fprintf(stderr, "\t-i 	case insensitive search\n");
	fprintf(stderr, "\t-n	print line numbers of matching files\n");
	fprintf(stderr, "\t-s	print the final distance score\n");
	fprintf(stderr, "\t-d	maximum distance allowed for match\n");
	exit(1);
}

/* convert a char array to lowercase, for case insensitive stuff */

void 
lowercase(char string[])
{
	static int  i;

	for (; string[i] != NULL; i++)
		string[i] = tolower(string[i]);
	return;
}

/*
 * jmatch is a small demo of edit distance functionality. sort of like
 * grep, but doesn't use regular expressions. note that whitespace
 * counts ...
 */

int
main(int argc, char *argv[])
{
	int i, ch, alg = LEVENSHTEIN_D;
	static int iflag, nflag, sflag, val;
	char target[1024];
	FILE *fin;

	while ((ch = getopt(argc, argv, "DHLd:hins")) != -1) {
		switch (ch) {
		case 'D':
			alg = DAMERAU_D;
			break;
		case 'H':
			alg = HAMMING_D;
			break;
		case 'L':
			alg = LEVENSHTEIN_D;
			break;
		case 'd':
			val = atoi(optarg);
			break;
		case 'i':
			iflag = 1;
			break;
		case 'n':
			nflag = 1;
			break;
		case 's':
			sflag = 1;
			break;
		case 'h':
		default:
			usage(argv[0]);
			break;
		}
	}

	if ((argc - optind) < 1) {
		printf("required arugments missing.\n");
		usage(argv[0]);
	}

	argv += optind;
	argc -= optind;

	strlcpy(target, argv[0], sizeof(target));

	if (iflag)
		lowercase(target);

	argv += 1;
	argc -= 1;

	/*
	 * loop over the files listed on the input and look for
 	 * matches ... if argv[0] is NULL (ie empty) open stdin and 
 	 * read from there, similar to how grep(1) works.
	 * use fgets() to parse a line at a time.
	 */

	for (i = 0; i <= argc; i++) {
		static int 	dobreak;	/* for stdin */

		/* XXX */
		if ((i > 0) && (i == argc))
			break;

		if (argv[0] == NULL) {
			fin = fdopen(STDIN_FILENO, "r");
			dobreak = 1;
		} else 
			fin = fopen(argv[i], "r");

		if (fin == NULL) {
			warn("%s", argv[i]);
			continue;
		}
			
		while (1) {
			char input[1024];
			static int 	lineno = 1;
			static int 	dist;

			fgets(input, sizeof(input), fin);
			if (feof(fin))
				break;

			if (iflag)
				lowercase(input);
	
			/* actually search the line */
			switch(alg) {
			case HAMMING_D:
				dist = hamming_d(target, strlen(target),
						 input, strlen(input));
				if (dist < 0)
					dist = 9999; 	/* should never match */
				break;
			case DAMERAU_D:
				dist = damerau_d(target, strlen(target),
						 input, strlen(input));
				break;
			default:
			case LEVENSHTEIN_D:
				dist = levenshtein_d(target, strlen(target), 
						     input, strlen(input));
				break;
			}

			/* print results if we're below tolerance threshhold */
			if (dist <= val) {
				if (argc > 1)
					printf("%s:", argv[i]);
				if (nflag)
					printf("%d:", lineno);
				if (sflag)
					printf("%d:", dist);
				printf("%s", input);
			}
			lineno++;
		}
		fclose(fin);

		if (dobreak)
			break;		/* XXX stdin was used ... cheap */
	}
	exit(0);
}
